"""
    Swagger Words
    ~~~~~~~~~~~~~

    Python friendly aliases to reserved Swagger words.

    :copyright: Copyright 2018 PlanGrid, Inc., see AUTHORS.
    :license: MIT, see LICENSE for details.
"""
from typing import Final

additional_properties: Final = "additionalProperties"
all_of: Final = "allOf"
allow_empty_value: Final = "allowEmptyValue"
any_of: Final = "anyOf"
api_key: Final = "apiKey"
array: Final = "array"
basic: Final = "basic"
binary: Final = "binary"
body: Final = "body"
boolean: Final = "boolean"
byte: Final = "byte"
collection_format: Final = "collectionFormat"
components: Final = "components"
consumes: Final = "consumes"
content: Final = "content"
csv: Final = "csv"
date: Final = "date"
date_time: Final = "date-time"
default: Final = "default"
definitions: Final = "definitions"
description: Final = "description"
double: Final = "double"
enum: Final = "enum"
external_docs: Final = "externalDocs"
exclusive_maximum: Final = "exclusiveMaximum"
exclusive_minimum: Final = "exclusiveMinimum"
explode: Final = "explode"
float_: Final = "float"
form: Final = "form"
format_: Final = "format"
header: Final = "header"
host: Final = "host"
in_: Final = "in"
info: Final = "info"
integer: Final = "integer"
int32: Final = "int32"
int64: Final = "int64"
items: Final = "items"
max_items: Final = "maxItems"
max_length: Final = "maxLength"
max_properties: Final = "maxProperties"
maximum: Final = "maximum"
min_items: Final = "minItems"
min_length: Final = "minLength"
min_properties: Final = "minProperties"
minimum: Final = "minimum"
multi: Final = "multi"
multiple_of: Final = "multipleOf"
name: Final = "name"
null: Final = "null"
nullable: Final = "x-nullable"
number: Final = "number"
oauth2: Final = "oauth2"
object_: Final = "object"
one_of: Final = "oneOf"
openapi: Final = "openapi"
operation_id: Final = "operationId"
parameters: Final = "parameters"
password: Final = "password"
path: Final = "path"
paths: Final = "paths"
pattern: Final = "pattern"
produces: Final = "produces"
properties: Final = "properties"
query: Final = "query"
ref: Final = "$ref"
request_body: Final = "requestBody"
required: Final = "required"
responses: Final = "responses"
schema: Final = "schema"
schemas: Final = "schemas"
schemes: Final = "schemes"
security: Final = "security"
security_definitions: Final = "securityDefinitions"
security_schemes: Final = "securitySchemes"
servers: Final = "servers"
simple: Final = "simple"
string: Final = "string"
style: Final = "style"
summary: Final = "summary"
swagger: Final = "swagger"
tags: Final = "tags"
title: Final = "title"
type_: Final = "type"
unique_items: Final = "uniqueItems"
url: Final = "url"
uuid: Final = "uuid"
variables: Final = "variables"
version: Final = "version"
