from flask_rebar.utils.request_utils import *  # noqa

"""
This file is here as a temporary import-only stub/shim in order to not break any existing imports
of the following form, e.g.,:
from flask_rebar.request_utils import marshal

When we release 2.0 and can make breaking changes, this stub should be removed as we are
consolidating "utility" modules such as request_utils.py into flask_rebar.utils
"""
